[!if=(FileExists, "FALSE")]
// [!CPPName] : Implementation of [!ClassName]

#include "stdafx.h"
#include "[!ProjectName].h"
#include "[!HeaderName]"
#include <algorithm>
[!else]
[!AddIncludeFile(TargetFile, "stdafx.h")]
[!AddStringToSymbol(ProjectName.h, ProjectName, ".h")]
[!AddIncludeFile(TargetFile, ProjectName.h)]
[!AddIncludeFile(TargetFile, HeaderName)]
[!AddIncludeFile(TargetFile, "algorithm")]
[!endif]
[!crlf]

/////////////////////////////////////////////////////////////////////////////
// [!ClassName]
[!crlf]
[!if=(ErrorInfoEnabled, "TRUE")]
STDMETHODIMP [!ClassName]::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_[!InterfaceName]
	};

	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}
[!endif]


[!crlf]
[!crlf]
/////////////////////////////////////////////////////////////////////////////
// Stream helpers
//   loadStream() -- loads data from a stream in dynamically allocated buffer
//   peekStream() -- just like loadStream() but doesn't change current read position
//
//  Both return dynamically allocated buffers.  Caller frees using operator delete[].
//  Both accept nRequestedSize which can be (size_t)-1 to get entire stream.
//  Both have an optional 3rd parameter for the number of bytes actual read.
//
static BYTE * loadStream ( IStream * pstm, size_t nRequestedSize, size_t & nActualSize )
{
    nActualSize = 0; // default value for output parameter
    if (! pstm) return 0;

    // determine the entire stream size, if no size is provided
    if (nRequestedSize == (size_t)-1)
    {
        STATSTG stat; 
        memset(&stat, 0, sizeof stat);
        if (FAILED(pstm->Stat(&stat, STATFLAG_NONAME))) return 0;
        nRequestedSize = (size_t)stat.cbSize.QuadPart;
    }

    // allocate a buffer and clear to zero
    BYTE * buffer = new unsigned char[2 + nRequestedSize];
    memset(buffer, 0, 2+nRequestedSize); // extra L'\0' in case being null-terminated is useful

    // read in the buffer and record how many bytes actually read
    ULONG nBytesRead = 0;
    HRESULT hr = pstm->Read(buffer, (ULONG)nRequestedSize, &nBytesRead);
    if (FAILED(hr))
    {
        delete[] buffer;
        buffer = 0;
    }

    // set the number of bytes actually read
    if (buffer)
        nActualSize = (size_t) nBytesRead;

    return buffer;
}

static BYTE * loadStream ( IStream * pstm, size_t nRequestedSize = (size_t) -1 /* defaults to entire stream size */ )
{
    size_t nUnused;
    return loadStream(pstm, nRequestedSize, nUnused);
}
[!crlf]
static BYTE * peekStream ( IStream * pstm, size_t nPeek, size_t & nPeeked )
{
    // load the requested number of bytes from the stream
    BYTE * buffer = loadStream(pstm, nPeek, nPeeked);

    // restore read position so the read is really a peek
    if (buffer && (nPeeked > 0))
    {
        LARGE_INTEGER nOffset; nOffset.QuadPart = - (LONGLONG)nPeeked;
        ATLASSERT(SUCCEEDED(pstm->Seek(nOffset, STREAM_SEEK_CUR, 0)));
    }

    return buffer;
}
[!crlf]
static BYTE * peekStream ( IStream * pstm, size_t nPeek = (size_t) -1 /* defaults to entire stream size */ )
{
    size_t nUnused;
    return peekStream(pstm, nPeek, nUnused);
}
[!crlf]
[!crlf]
/////////////////////////////////////////////////////////////////////////////
// ATLTRACE helpers
//   _TRACER(x) -- returns a non-null string value for various types x.
//                 The return type is either BSTR or CComBSTR.
//   _AtlTraceBigBufer() -- a replacement for ATL's ATLTRACE mechanism
//                 which has a limit of 512 chars and fires an assert
//                 if the message is greater than 512.  Firing an 
//                 assert message box inside a component loaded into a 
//                 service is not good.
#ifdef _DEBUG
#undef ATLTRACE
#define ATLTRACE _AtlTraceBigBuffer
[!crlf]
void _AtlTraceBigBuffer ( LPCTSTR pszFormat, ... )
{
	const size_t nSize = 8192;
	TCHAR * sz = new TCHAR[nSize];
	memset(sz, 0, nSize);
	va_list args;
	va_start(args, pszFormat);
	_vsntprintf(sz, nSize-1, pszFormat, args);
    va_end(args);
	OutputDebugString(sz); //would call ATLTRACE but szReg is > 512 bytes
	delete[] sz;
}
static BSTR _TRACER ( BSTR s )
{
    return s ? s : L"";
}
[!crlf]
static CComBSTR _TRACER ( IStream * pstm, size_t nPrefixLen = 50 ) // nPrefixLen = how many bytes of the stream prefix should be returned
{
    if (! pstm) return L"<null>";
    wchar_t wsz[4096];
    wcscpy(wsz, L"{size:");
    STATSTG stat;
    memset(&stat, 0, sizeof stat);
    if (SUCCEEDED(pstm->Stat(&stat, STATFLAG_NONAME)))
        swprintf(wsz+wcslen(wsz), L"%I64u", stat.cbSize.QuadPart);
    else
        wcscat(wsz, L"<unavailable>");
    if (nPrefixLen)
    {
        wcscat(wsz, L" data:");
        // peek at data; get 1 more byte than nPrefixLen so can test if there is more data
        size_t nBytesRead;
        char * pszData = (char *)peekStream(pstm, nPrefixLen+1, nBytesRead);
        bool bMoreData = (nBytesRead > nPrefixLen);
        if (bMoreData) -- nBytesRead; // don't use the extra byte read just for testing purposes
        if (! pszData)
            wcscat(wsz, L"<unavailable>");
        else
        {
            wcscat(wsz, L"\"");
            for ( int i = 0;  i < (int)nBytesRead;  ++i )
                if (isprint((unsigned char)pszData[i]))
                    swprintf(wsz+wcslen(wsz), L"%C", pszData[i]);
                else
                    switch (pszData[i])
                    {
                        case '\0': wcscat(wsz, L"\\0"); break;
                        case '\r': wcscat(wsz, L"\\r"); break;
                        case '\n': wcscat(wsz, L"\\n"); break;
                        case '\t': wcscat(wsz, L"\\t"); break;
                        case '\"': wcscat(wsz, L"\\\""); break;
                        default:   swprintf(wsz+wcslen(wsz), L"\\x%02X", (int)(unsigned char)pszData[i]); break;
                    }
            if (bMoreData) wcscat(wsz, L"...");
            wcscat(wsz, L"\"");
        }
        delete[] (BYTE*)pszData;
    }
    wcscat(wsz, L"}");
    return wsz;
}
[!crlf]
static CComBSTR _TRACER ( IDictionary * pdict )
{
    if (! pdict) return L"<null>";
    // get the number of items in the dictionary
    wchar_t wsz[8192];
    wcscpy(wsz, L"{count:");
    long nCount = 0;
    if (FAILED(pdict->get_Count(&nCount)))
        wcscat(wsz, L"<unavailable>");
    else
    {
        swprintf(wsz+wcslen(wsz), L"%ld", nCount);
        if (nCount > 0) 
        {
            wcscat(wsz, L" items:");
            CComPtr<IUnknown> spunkEnum;
            if (FAILED(pdict->get__NewEnum(&spunkEnum))) spunkEnum = 0;
            CComQIPtr<IEnumVARIANT> spEnum(spunkEnum);
            if (! spEnum)
                wcscat(wsz, L"<unavailable>");
            else
                for ( ;; )
                {
                    CComVariant vaKey, vaValue;
                    if (S_OK != spEnum->Next(1, &vaKey, 0)) break;
                    if (FAILED(vaKey.ChangeType(VT_BSTR))) continue;
                    swprintf(wsz+wcslen(wsz), L"[%ls]=", vaKey.bstrVal ? vaKey.bstrVal : L"");
                    if (SUCCEEDED(pdict->get_Value(vaKey.bstrVal, &vaValue)) && SUCCEEDED(vaValue.ChangeType(VT_BSTR)))
                        swprintf(wsz+wcslen(wsz), L"\"%ls\" ", vaValue.bstrVal ? vaValue.bstrVal : L"");
                    else
                        wcscat(wsz, L"<unavailable>");
                }
        }
    }
    wcscat(wsz, L"}");
    return wsz;
}
#endif
[!crlf]
[!crlf]

[!crlf]
/////////////////////////////////////////////////////////////////////////////
// IBizTalkParserComponent implementation
//

[!crlf]
// IBizTalkParserComponent::ProbeInterchangeFormat()
// -------------------------------------------------
//  identify whether the given data stream matches the format handled by this parser.
//
//  Called before any other member of this interface.  No other interface member is called
//  if this function indicates the data stream is not handled by the parser.
//
//  on input,
//      pstmData = stream containing the data for the parser.  Read-only.  Only members
//                 Read(), Seek(), and Stat() are defined.  The current position in
//                 the stream is not necessarily zero.
//      bFromFile = TRUE, indicates pstmData was loaded from a file and no assumption can be made about format
//                  FALSE, indicates pstmData was initialized from a BSTR which indicates UNICODE format
//      bstrEnvName = name of the associated envelope or null if there is none.  Is the value
//                    passed into IInterchange (see below)
//      pstmReceiptData = stream containing the receipt document
//
//      The first three parameters are derived from the parameters to IInterchange::Submit() 
//      and IInterchange::SubmitSync().  Those two functions are the only way to get an
//      an interchange submitted to BizTalk.  IInterchange accepts the data for the interchange
//      as a BSTR or a file reference.  pstmData is an IStream on either data source and 
//      bFromFile indicates which kind of data source (file or string).
//
//  on output,
//      pbstrFormat = document format.  Valid for IBizTalkDocument::Type property which must have
//                    one of the values "x12", "edifact", "custom xml", "custom", "flatfile", or "reliable".
//                    This value is set to null if S_FALSE is returned.
//
//  return value
//      S_OK, if the format is handled by this parser.  pbstrFormat is set.
//      S_FALSE, if the format is not handled by this parser.  pbstrFormat is set to null.
//      E_POINTER if input parameter pointers are null
// 
//  implementation note:
//      Per the BizTalk documention, pstmData is addrefed and stored if S_OK is returned.
//
STDMETHODIMP [!ClassName]::ProbeInterchangeFormat (IStream * pstmData, BOOL bFromFile, BSTR bstrEnvName, IStream * pstmReceiptData, BSTR * pbstrFormat)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::ProbeInterchangeFormat([in] pstmData=%ls, [in] bFromFile=%ld, [in] bstrEnvName=\"%ls\", [in] pstmReceiptData=%ls)\n"),
             (BSTR)_TRACER(pstmData), bFromFile, (BSTR)_TRACER(bstrEnvName), (BSTR)_TRACER(pstmReceiptData));
    if (!pstmData || !pbstrFormat) 
    {
        ATLTRACE("\nExit IBizTalkParserComponent::ProbeInterchangeFormat(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }

[!crlf]
    // acquire the biztalk configuration interface
    HRESULT hr = m_spconfig.CoCreateInstance(L"BizTalk.BizTalkConfig");
    ATLASSERT(SUCCEEDED(hr) && m_spconfig.p);
    if (FAILED(hr)) return E_FAIL;

[!crlf]
    // acquire the biztalk envelope, if a name is provided
    CComQIPtr<IBizTalkEnvelope> spenvelope;
    if (bstrEnvName && *bstrEnvName) // envelope name is provided
    {
        CComPtr<IDispatch> spdisp;
        hr = m_spconfig->CreateEnvelope(&spdisp); // create empty envelope
        ATLASSERT(SUCCEEDED(hr) && spdisp.p);
        spenvelope = spdisp;
        ATLASSERT(spenvelope.p);
        hr = spenvelope->LoadByName(bstrEnvName); // load the envelope
        ATLASSERT(SUCCEEDED(hr));
    }

[!crlf]
#pragma message("TBD: Custom Parser:  Decide whether the given stream can be handled by this parser in IBizTalkParserComponent::ProbeInterchangeFormat()")
    // Decide whether can handle this format and set local variable bCanHandleFormat accordingly.
    //  The stock implementation here tests if the stream begins with a schema identifier.
    //  Care is taken to not change the current stream position.
    const char SCHEMAID[] = "insert your schema identifier here";
    size_t nSchemaLen = sizeof SCHEMAID - 1, nPeeked = 0;
    char * buffer = (char*)::peekStream(pstmData, nSchemaLen, nPeeked);
    bool bCanHandleFormat = buffer && (nPeeked == nSchemaLen)
                            && (0 == ::memcmp(buffer, SCHEMAID, nSchemaLen));

[!crlf]
    // return whether this parser recognizes the format
    // if true is returned, then a copy of the data stream is retained for subsequent use
    if (! bCanHandleFormat)
    {
        *pbstrFormat = 0;
        ATLTRACE("\nExit IBizTalkParserComponent::ProbeInterchangeFormat([out] pbstrFormat=%ls), return S_FALSE\n", _TRACER(*pbstrFormat));
        return S_FALSE;
    }
[!crlf]
    // TBD:  select different format if "custom" is not appropriate
    *pbstrFormat = CComBSTR(L"custom").Detach();
[!crlf]
    // cache the stream for the interchange for later use
    m_spstmData = pstmData;
    ATLTRACE("\nExit IBizTalkParserComponent::ProbeInterchangeFormat([out] pbstrFormat=\"%ls\"), return S_OK\n", _TRACER(*pbstrFormat));
    return S_OK;
}

[!crlf]
// IBizTalkParserComponent::GetInterchangeDetails()
// ------------------------------------------------
//  receive information about the interchange such as the envelope name, if any,
//  and identifiers for the source and destination organizations.
//
//  on input,
//      pdict = dictionary of <field,value> pairs.  See the BizTalk documentation 
//              for the fields stored in the dictionary.
// 
STDMETHODIMP [!ClassName]::GetInterchangeDetails (IDictionary * pdict)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GetInterchangeDetails([in] pdict=%ls)\n"), (BSTR)_TRACER(pdict));
    if (! pdict) 
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GetInterchangeDetails(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }
#pragma message("TBD: Custom Parser:  Record interchange details in IBizTalkParserComponent::GetInterchangeDetails()")
    ATLTRACE(_T("\nExit IBizTalkParserComponent::GetInterchangeDetails(), return S_OK\n"));
    return S_OK;
}

[!crlf]
// IBizTalkParserComponent::GroupsExist()
// --------------------------------------
//  flag whether the interchange contains groups.  Called before any 
//  parsing takes place.
//
//  An interchange can be a single document, a sequence of documents, or 
//  a sequence of groups.  A group itself contains a sequence of documents.
//
//  Called before any parsing takes place, immediately after ProbeInterchangeFormat().
//
//  on output,
//      pbGroupsExist = whether there are groups
// 
STDMETHODIMP [!ClassName]::GroupsExist (BOOL * pbGroupsExist)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GroupsExist()\n"));
    if (! pbGroupsExist) 
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GroupsExist(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }
#pragma message("TBD: Custom Parser:  Flag whether there are groups in IBizTalkParserComponent::GroupsExist()")
    *pbGroupsExist = FALSE;
    ATLTRACE(_T("\nExit IBizTalkParserComponent::GroupsExist([out] pbGroupsExist=%ld), return S_OK\n"), *pbGroupsExist);
    return S_OK;
}

[!crlf]
// IBizTalkParserComponent::GetGroupDetails()
// ------------------------------------------
//  specifies details for a single group to be used in the BizTalk Tracking Database.
//  Called before each group gets processed.
//
//  on input,
//      pdict = dictionary of <field,value> pairs.  See the article
//              "Supporting the Tracking Database with Parser and Serializer Components"
//              in the BizTalk documentation for the fields stored in the dictionary.
// 
STDMETHODIMP [!ClassName]::GetGroupDetails (IDictionary * pdict)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GetGroupDetails([in] pdict=%ls)\n"), (BSTR)_TRACER(pdict));
    if (! pdict) 
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GetGroupDetails(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }
#pragma message("TBD: Custom Parser:  Specify group details in IBizTalkParserComponent::GetGroupDetails()")
    ATLTRACE(_T("\nExit IBizTalkParserComponent::GetGroupDetails([out] pdict=%ls), return S_OK\n"), (BSTR)_TRACER(pdict));
    return S_OK;
}

[!crlf]
// IBizTalkParserComponent::GetGroupSize()
// ---------------------------------------
//  specify group-level information for the BizTalk Tracking database.  Called after
//  the last document in a group has been parsed.
//
//  on output,
//      pnGroupSize = total size of the group
//      pbLastGroup = whether this is the last group.  FALSE means another group will be parsed.
// 
STDMETHODIMP [!ClassName]::GetGroupSize (long * pnGroupSize, BOOL * pbLastGroup)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GetGroupSize()\n"));
    if (!pnGroupSize || !pbLastGroup) 
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GetGroupSize(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }
#pragma message("TBD: Custom Parser:  Specify group size in IBizTalkParserComponent::GetGroupSize()")
    *pnGroupSize = 0;
    *pbLastGroup = TRUE;
    ATLTRACE(_T("\nExit IBizTalkParserComponent::GetGroupSize([out] pnGroupSize=%ld, [out] pbLastGroup=%ld), return E_NOTIMPL\n"), *pnGroupSize, *pbLastGroup);
    return E_NOTIMPL;
}

[!crlf]
// IBizTalkParserComponent::GetNextDocument()
// ------------------------------------------
//  parse the next document from the data stream.  Recall that the data stream was stored
//  as a class member in the call to ProbeInterchangeFormat().
//
//  on input,
//      pdict = dictionary into which to place results
//      bstrDocName = name of the document specification
//
//  on output,
//      pdict is updated.  Field "WORKING_DATA " contains the XML for the parsed document.
//                         Sets fields for routing if the document is self-routing.  These fields
//                         are "SrcQual", "SrcID", "DestQual", and "DestID".  
//                         Recall:  Key names are not case sensitive.
//      pbDocIsValid = whether the document instance is valid.  If false is returned, the read position
//                     on the stream must be advanced to the start of the next document.  If this error
//                     recovery is not possible, then E_FAIL must be returned to abort the interchange.
//      pbLastDocument = whether this doucment is the last in the current group or the entire interchange
//      pnReceiptGenerated = receipt indicator.
//      pbDocIsReceipt = whether the document is a receipt
//      pbstrCorrelationCompProgID = prog id for the correlation of documents
//
//  return value
//      S_OK -- document was parsed or if an error was detected, the stream position was advanced
//              to the start of the next document
//      E_FAIL -- if parsing the document failed and the interchange should be aborted
//
STDMETHODIMP [!ClassName]::GetNextDocument (IDictionary * pdict, BSTR bstrDocName, BOOL * pbDocIsValid, BOOL * pbLastDocument, enum GeneratedReceiptLevel * pnReceiptGenerated, BOOL * pbDocIsReceipt, BSTR * pbstrCorrelationCompProgID)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GetNextDocument([in] pdict=%ls, [in] bstrDocName=\"%ls\")\n"), (BSTR)_TRACER(pdict), (BSTR)_TRACER(bstrDocName));
    if (!pdict || !pbDocIsValid || !pbLastDocument || !pnReceiptGenerated || !pbDocIsReceipt || !pbstrCorrelationCompProgID)
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GetNextDocument(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }

[!crlf]
    // acquire the biztalk document specification, if a name is provided
    CComQIPtr<IBizTalkDocument> spdocument;
    if (bstrDocName && *bstrDocName) // document specification name is provided
    {
        CComPtr<IDispatch> spdisp;
        HRESULT hr = m_spconfig->CreateDocument(&spdisp); // create empty document specification
        ATLASSERT(SUCCEEDED(hr) && spdisp.p);
        spdocument = spdisp;
        ATLASSERT(spdocument.p);
        hr = spdocument->LoadByName(bstrDocName); // load the document specification
        ATLASSERT(SUCCEEDED(hr));
    }

[!crlf]
#pragma message("TBD: Custom Parser:  Parse next document in IBizTalkParserComponent::GetNextDocument()")
    pdict->put_Value(L"working_data", CComVariant(L"<xml/>"));
    *pbDocIsValid = TRUE;
    *pbLastDocument = TRUE;
    *pnReceiptGenerated = NoReceiptGenerated;
    *pbDocIsReceipt = FALSE;
    *pbstrCorrelationCompProgID = 0;
    ATLTRACE(_T("\nExit IBizTalkParserComponent::GetNextDocument([out] pdict=%ls, [out] pbDocIsValid=%ld, [out] pbLastDocument=%ld, [out] pnReceiptGenerated=%ld, [out] pbDocIsReceipt=%ld, [out] pbstrCorrelationCompProgID=\"%ls\"), return S_OK\n"), 
                (BSTR)_TRACER(pdict), *pbDocIsValid, *pbLastDocument, (long)*pnReceiptGenerated, *pbDocIsReceipt, (BSTR)_TRACER(*pbstrCorrelationCompProgID));
    return S_OK;
}

[!crlf]
// IBizTalkParserComponent::GetNativeDocumentOffsets ()
// ----------------------------------------------------
//  specify the size and position of the unparsed data for the last document parsed
//  The requirement here is that the unparsed data be a contiguous block in the stream.
//
//  In contrast, the groups do not have to be contiguous and documents within a group
//  do not have to appear in any particular order.  The parsing has the opportunity to
//  reorganize the input stream into groups and documents within each group.  However, the
//  data for each each document must be contiguous.
//
//  on output,
//      pbSizeFromXMLDoc = TRUE,  if BizTalk should infer the size and position.  In this case,
//                                the other 2 output parameters are not used.
//                         FALSE, if the size and position are specifed in the other 2 output parameters.
//
//                         Inference is done based on the XML returned from GetNextDocument().
//      pnStartOffset = the offset within the stream for the start of the document's data in the unparsed form
//                      set only if pbSizeFromXMLDoc is set to FALSE
//      pnDocLength = the size within the stream for the document's data
//                    set only if pbSizeFromXMLDoc is set to FALSE
//
STDMETHODIMP [!ClassName]::GetNativeDocumentOffsets (BOOL * pbSizeFromXMLDoc, LARGE_INTEGER * pnStartOffset, long * pnDocLength)
{
    ATLTRACE(_T("\nEnter IBizTalkParserComponent::GetNativeDocumentOffsets()\n"));
    if (!pbSizeFromXMLDoc || !pnStartOffset || !pnDocLength)
    {
        ATLTRACE("\nExit IBizTalkParserComponent::GetNativeDocumentOffsets(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }
    *pbSizeFromXMLDoc = TRUE; // let BizTalk calculate
#pragma message("TBD: Custom Parser:  Specify size and position in IBizTalkParserComponent::GetNativeDocumentOffsets()")
    #if _DEBUG
        ATLTRACE(_T("\nExit IBizTalkParserComponent::GetNativeDocumentOffsets([out] pbSizeFromXMLDoc=%ld"), *pbSizeFromXMLDoc);
        if (! *pbSizeFromXMLDoc)
            ATLTRACE(_T(", [out] pnStartOffset=%I64d, [out] pnDocLength=%ld"), pnStartOffset->QuadPart, *pnDocLength);
        ATLTRACE(_T(")\n"));
    #endif
    return S_OK;
}
