// FormPostSerializer.cpp : Implementation of CFormPostSerializer
#include "stdafx.h"
#include "FormPostBatchSerializer.h"
#include "FormPostSerializer.h"
#include <ocidl.h> // for IPicture
#include <time.h>
#import "msxml3.dll" rename_namespace("XMLDOM"), raw_interfaces_only

/////////////////////////////////////////////////////////////////////////////
// CFormPostSerializer
//


/////////////////////////////////////////////////////////////////////////////
// Stream helpers
//   loadStream() -- loads data from a stream in dynamically allocated buffer
//   peekStream() -- just like loadStream() but doesn't change current read position
//
//  Both return dynamically allocated buffers.  Caller frees using operator delete[].
//  Both accept nRequestedSize which can be (size_t)-1 to get entire stream.
//  Both have an optional 3rd parameter for the number of bytes actual read.
//
static BYTE * loadStream ( IStream * pstm, size_t nRequestedSize, size_t & nActualSize )
{
    nActualSize = 0; // default value for output parameter
    if (! pstm) return 0;

    // determine the entire stream size, if no size is provided
    if (nRequestedSize == (size_t)-1)
    {
        STATSTG stat; 
        memset(&stat, 0, sizeof stat);
        if (FAILED(pstm->Stat(&stat, STATFLAG_NONAME))) return 0;
        nRequestedSize = (size_t)stat.cbSize.QuadPart;
    }

    // allocate a buffer and clear to zero
    BYTE * buffer = new unsigned char[2 + nRequestedSize];
    memset(buffer, 0, 2+nRequestedSize); // extra L'\0' in case being null-terminated is useful

    // read in the buffer and record how many bytes actually read
    ULONG nBytesRead = 0;
    HRESULT hr = pstm->Read(buffer, (ULONG)nRequestedSize, &nBytesRead);
    if (FAILED(hr))
    {
        delete[] buffer;
        buffer = 0;
    }

    // set the number of bytes actually read
    if (buffer)
        nActualSize = (size_t) nBytesRead;
    return buffer;
}
static BYTE * loadStream ( IStream * pstm, size_t nRequestedSize = (size_t) -1 /* defaults to entire stream size */ )
{
    size_t nUnused;
    return loadStream(pstm, nRequestedSize, nUnused);
}

static BYTE * peekStream ( IStream * pstm, size_t nPeek, size_t & nPeeked )
{
    // load the requested number of bytes from the stream
    BYTE * buffer = loadStream(pstm, nPeek, nPeeked);

    // restore read position so the read is really a peek
    if (buffer && (nPeeked > 0))
    {
        LARGE_INTEGER nOffset; nOffset.QuadPart = - (LONGLONG)nPeeked;
        ATLASSERT(SUCCEEDED(pstm->Seek(nOffset, STREAM_SEEK_CUR, 0)));
    }
    return buffer;
}

static BYTE * peekStream ( IStream * pstm, size_t nPeek = (size_t) -1 /* defaults to entire stream size */ )
{
    size_t nUnused;
    return peekStream(pstm, nPeek, nUnused);
}


/////////////////////////////////////////////////////////////////////////////
// ATLTRACE helpers
//   _TRACER(x) -- returns a non-null string value for various types x.
//                 The return type is either BSTR or CComBSTR.
//   _AtlTraceBigBufer() -- a replacement for ATL's ATLTRACE mechanism
//                 which has a limit of 512 chars and fires an assert
//                 if the message is greater than 512.  Firing an 
//                 assert message box inside a component loaded into a 
//                 service is not good.
#ifdef _DEBUG
#undef ATLTRACE
#define ATLTRACE _AtlTraceBigBuffer
void _AtlTraceBigBuffer ( LPCTSTR pszFormat, ... )
{
    const size_t nSize = 8192;
    TCHAR * sz = new TCHAR[nSize];
    memset(sz, 0, nSize);
    va_list args;
    va_start(args, pszFormat);
    _vsntprintf(sz, nSize-1, pszFormat, args);
    va_end(args);
    OutputDebugString(sz); //would call ATLTRACE but szReg is > 512 bytes
    delete[] sz;
}
static BSTR _TRACER ( BSTR s )
{
    return s ? s : L"";
}
static CComBSTR _TRACER ( IStream * pstm, size_t nPrefixLen = 50 ) // nPrefixLen = how many bytes of the stream prefix should be returned
{
    if (! pstm) return L"<null>";
    wchar_t wsz[4096];
    wcscpy(wsz, L"{size:");
    STATSTG stat;
    memset(&stat, 0, sizeof stat);
    if (SUCCEEDED(pstm->Stat(&stat, STATFLAG_NONAME)))
        swprintf(wsz+wcslen(wsz), L"%I64u", stat.cbSize.QuadPart);
    else
        wcscat(wsz, L"<unavailable>");
    if (nPrefixLen)
    {
        wcscat(wsz, L" data:");
        // peek at data; get 1 more byte than nPrefixLen so can test if there is more data
        size_t nBytesRead;
        char * pszData = (char *)peekStream(pstm, nPrefixLen+1, nBytesRead);
        bool bMoreData = (nBytesRead > nPrefixLen);
        if (bMoreData) -- nBytesRead; // don't use the extra byte read just for testing purposes
        if (! pszData)
            wcscat(wsz, L"<unavailable>");
        else
        {
            wcscat(wsz, L"\"");
            for ( int i = 0;  i < (int)nBytesRead;  ++i )
                if (isprint((unsigned char)pszData[i]))
                    swprintf(wsz+wcslen(wsz), L"%C", pszData[i]);
                else
                    switch (pszData[i])
                    {
                        case '\0': wcscat(wsz, L"\\0"); break;
                        case '\r': wcscat(wsz, L"\\r"); break;
                        case '\n': wcscat(wsz, L"\\n"); break;
                        case '\t': wcscat(wsz, L"\\t"); break;
                        case '\"': wcscat(wsz, L"\\\""); break;
                        default:   swprintf(wsz+wcslen(wsz), L"\\x%02X", (int)(unsigned char)pszData[i]); break;
                    }
            if (bMoreData) wcscat(wsz, L"...");
            wcscat(wsz, L"\"");
        }
        delete[] (BYTE*)pszData;
    }
    wcscat(wsz, L"}");
    return wsz;
}

static CComBSTR _TRACER ( IDictionary * pdict )
{
    if (! pdict) return L"<null>";
    // get the number of items in the dictionary
    wchar_t wsz[8192];
    wcscpy(wsz, L"{count:");
    long nCount = 0;
    if (FAILED(pdict->get_Count(&nCount)))
        wcscat(wsz, L"<unavailable>");
    else
    {
        swprintf(wsz+wcslen(wsz), L"%ld", nCount);
        if (nCount > 0) 
        {
            wcscat(wsz, L" items:");
            CComPtr<IUnknown> spunkEnum;
            if (FAILED(pdict->get__NewEnum(&spunkEnum))) spunkEnum = 0;
            CComQIPtr<IEnumVARIANT> spEnum(spunkEnum);
            if (! spEnum)
                wcscat(wsz, L"<unavailable>");
            else
                for ( ;; )
                {
                    CComVariant vaKey, vaValue;
                    if (S_OK != spEnum->Next(1, &vaKey, 0)) break;
                    if (FAILED(vaKey.ChangeType(VT_BSTR))) continue;
                    swprintf(wsz+wcslen(wsz), L"[%ls]=", vaKey.bstrVal ? vaKey.bstrVal : L"");
                    if (SUCCEEDED(pdict->get_Value(vaKey.bstrVal, &vaValue)) && SUCCEEDED(vaValue.ChangeType(VT_BSTR)))
                        swprintf(wsz+wcslen(wsz), L"\"%ls\" ", vaValue.bstrVal ? vaValue.bstrVal : L"");
                    else
                        wcscat(wsz, L"<unavailable>");
                }
        }
    }
    wcscat(wsz, L"}");
    return wsz;
}
#endif



/////////////////////////////////////////////////////////////////////////////
// IBizTalkSerializer implementation
//

// IBizTalkSerializer::Init()
// --------------------------
//  receive the output stream to use during serializing.  Also, receive the number of
//  documents in the interchange.  For BizTalk 2002, this number is always 1.
//
//  Called before any other member of this interface.
//
//  on input,
//      bstrSourceQualifier = qualifier part of the source organization
//      bstrSourceID = identifier part of the source organization
//      bstrDestinationQualifier = qualifier part of the destination organization
//      bstrDestinationID = identifier part of the destination organization
//      nEnvelopeID = identifier for the envelope used in the channel
//      pdictDelimiters = dictionary containing any delimiters used for serialization
//      pstmOutput = stream into which to write all output
//      nNumDocs = number of documents in the interchange.  Will always be 1 in BizTalk 2002.
//      nPortID = identifier for the port to which this channel using this serializer is connected
//
//  return value
//      is S_OK on success and a negative HRESULT otherwise
//
//  implementation note:
//      Per the BizTalk documention, pstmOutput is addrefed and stored.
//
STDMETHODIMP CFormPostSerializer::Init(BSTR bstrSourceQualifier, BSTR bstrSourceID, BSTR bstrDestinationQualifier, BSTR bstrDestinationID,
                                       long nEnvelopeID, IDictionary * pdictDelimiters, IStream * pstmOutput, long nNumDocs, long nPortID)
{
    ATLTRACE(_T("\nEnter IBizTalkSerializer::Init([in] bstrSourceQualifier=\"%ls\", [in] bstrSourceID=\"%ls\", [in] bstrDestinationQualifier=\"%ls\", [in] bstrDestinationID=\"%ls\", ")
             _T("[in] nEnvelopeID=%ld, [in] pdictDelimiters=%ls, [in] nNumDocs=%ld, [in] nPortID=%ld)\n"),
             (BSTR)_TRACER(bstrSourceQualifier), (BSTR)_TRACER(bstrSourceID), (BSTR)_TRACER(bstrDestinationQualifier), (BSTR)_TRACER(bstrDestinationID), 
             nEnvelopeID, (BSTR)_TRACER(pdictDelimiters), nNumDocs, nPortID);
    if (!pdictDelimiters || !pstmOutput) 
    {
        ATLTRACE("\nExit IBizTalkSerializer::Init(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }

    // in BizTalk 2002 release, number of documents should be exactly 1
    if (nNumDocs != 1) 
    {
        ATLTRACE("\nExit IBizTalkSerializer::Init(), return E_NOTIMPL\n");
        ATLASSERT(0);
        return E_NOTIMPL;
    }

    // cache the output stream for later use
    m_spstmOutput = pstmOutput;
    ATLTRACE("\nExit IBizTalkSerializer::Init(), return S_OK\n");
    return S_OK;
}


CComPtr<IPicture> makeReceiptBitmap ( LPCTSTR pszReceiptText )
{
    HDC hdcdisplay = CreateDC(_T("DISPLAY"), 0, 0, 0); // get dc for screen
    RECT rcText = { 0, 0, 0, 0 };
    DrawText(hdcdisplay, pszReceiptText, -1, &rcText, DT_CALCRECT); // calc size of rect for given text
    RECT rc = { 0, 0, rcText.right+24, rcText.bottom+24 };   // expand this rect by a 12px margin
    OffsetRect(&rcText, 12, 12); // let rect for text have a 12px margin
    HDC hdcmem     = CreateCompatibleDC(hdcdisplay); // get mem dc matching the display dc
    HBITMAP hbm    = CreateCompatibleBitmap(hdcdisplay, rc.right, rc.bottom); // create a bitmap for use with the mem dc
    HBITMAP hbmold = (HBITMAP) SelectObject(hdcmem, hbm); // select the bitmap into the mem dc so can write to bitmap
    WORD pixelpattern[8] = { 0xAAAA, 0x5555, 0xAAAA, 0x5555, 0xAAAA, 0x5555, 0xAAAA, 0x5555 }; 
                    // note: 0xAAAA = 1010101010101010   pixel pattern
                    //       0x5555 = 0101010101010101   shifted pixel pattern
    HBITMAP hbmHalftone = CreateBitmap(8, 8, 1, 1, pixelpattern); // create half-tone bitmap
    HBRUSH hbrHalftone = CreatePatternBrush(hbmHalftone); // create half-tone brush
    HBRUSH hbrSolid = CreateSolidBrush(RGB(128,64,0)); // create brush for main brackground
    FillRect(hdcmem, &rc, hbrHalftone); // fill with half-tone
    InflateRect(&rc, -6, -6); // reduce by 6 px margin on all sides
    FillRect(hdcmem, &rc, hbrSolid); // fill interior rect by with solid brush
    DrawEdge(hdcmem, &rc, EDGE_BUMP, BF_RECT); // place an edge around interior fill
    InflateRect(&rc, 6, 6); // restore rect to full size
    DrawEdge(hdcmem, &rc, EDGE_BUMP, BF_RECT); // place an edge around full rect
    DeleteObject(hbmHalftone); // done with half-tone bitmap
    DeleteObject(hbrHalftone); // done with half-tone brush
    DeleteObject(hbrSolid);    // done with solid brush
    SetBkMode(hdcmem, TRANSPARENT);
    SetTextColor(hdcmem, RGB(255,255,255));
    DrawText(hdcmem, pszReceiptText, -1, &rcText, 0); // place the text on the bitmap, interior rect
    CComPtr<IPicture> sppict; 
    PICTDESC pd = { sizeof pd, PICTYPE_BITMAP };
    pd.bmp.hpal = 0; pd.bmp.hbitmap = hbm; // create picture object on this bitmap
    if (FAILED(OleCreatePictureIndirect(&pd, IID_IPicture, TRUE/*own the HBITMAP*/, (void**)&sppict)))
        sppict = 0;
    SelectObject(hdcmem, hbmold); // restore original setting (required step)
    DeleteDC(hdcdisplay); // done with dc for the screen
    DeleteDC(hdcmem);     // done with memory dc
    return sppict;
}


CComVariant xmllookup ( CComPtr<XMLDOM::IXMLDOMDocument2> spxml, const wchar_t * pwszXPath, VARTYPE vt, const CComVariant & vaDefault )
{
    CComVariant vaValue;
    CComPtr<XMLDOM::IXMLDOMNode> spnode;
    if (spxml
        && SUCCEEDED(spxml->selectSingleNode(const_cast<wchar_t*>(pwszXPath), &spnode))
        && spnode
        && SUCCEEDED(spnode->get_nodeTypedValue(&vaValue))
        && SUCCEEDED(vaValue.ChangeType(vt)))
        return vaValue;
    return vaDefault;
}


// IBizTalkSerializer::AddDocument()
// ---------------------------------
//  serialize the given document to the output stream passed into the 
//  this serializer by the Init() function.
//
//  on input,
//      nDocHandle = an integer handle to associated with this document.
//                   When documents are grouped using GetGroupInfo() and GetDocInfo(),
//                   a document will be identified by its handle.
//      pdictTransport = the transport dictionary containing the the XML document to
//                       serialize under the dictionary key "working_data".  Remember
//                       that keys in IDictionary are not case sensitive.
//      bstrTrackID = id used in the tracking database for this document
//      nChannelID = id of the channel in which this custom serializer is being used
//
//  return value
//      is S_OK on success and a negative HRESULT otherwise
//
STDMETHODIMP CFormPostSerializer::AddDocument(long nDocHandle, IDictionary * pdictTransport, BSTR bstrTrackID, long nChannelID)
{
    ATLTRACE(_T("\nEnter IBizTalkSerializer::AddDocument([in] nDocHandle=%ld, [in] pdictTransport=%ls, [in] bstrTrackID=\"%ls\", [in] nChannelID=%ld)\n"), 
                  nDocHandle, (BSTR)_TRACER(pdictTransport), (BSTR)_TRACER(bstrTrackID), nChannelID);
    if (! pdictTransport) 
    {
        ATLTRACE("\nExit IBizTalkSerializer::AddDocument(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }

    // get the XML document from the transport dictionary
    CComPtr<XMLDOM::IXMLDOMDocument2> spxml;
    VARIANT_BOOL bLoaded;
    CComVariant vaKey(L"working_data"), vaDocument;
    HRESULT hr = pdictTransport->get_Value(vaKey.bstrVal, &vaDocument);
    if (SUCCEEDED(hr)) hr = vaDocument.ChangeType(VT_BSTR);
    if (SUCCEEDED(hr)) hr = spxml.CoCreateInstance(__uuidof(XMLDOM::DOMDocument30));
    if (SUCCEEDED(hr)) hr = spxml->loadXML(vaDocument.bstrVal, &bLoaded);
    if (FAILED(hr) || (bLoaded == VARIANT_FALSE))
    {
        ATLTRACE("\nExit IBizTalkSerializer::AddDocument(), return E_FAIL\n");
        ATLASSERT(0);
        return E_FAIL;
    }

    // extract the fields from the XML
    CComVariant vaBroker  = xmllookup(spxml, L"//broker",    VT_BSTR, L"???");
    CComVariant vaStock   = xmllookup(spxml, L"//stock",     VT_BSTR, L"???");
    CComVariant vaQuanity = xmllookup(spxml, L"//quantity",  VT_I4,   0l);
    CComVariant vaPrice   = xmllookup(spxml, L"//min-price", VT_R4,   0.0f);

    // format the text to go on the image
    time_t tm = time(0);
    TCHAR sz[1024];
    _stprintf(sz, _T("Stock Trade Receipt\n\n")
                  _T("     %ls\n")
                  _T("Broker:  %ls\nStock:  %ls\n\n")
                  _T("%d share%s at $%0.2f"),
              _tasctime(localtime(&tm)), vaBroker.bstrVal, vaStock.bstrVal, vaQuanity.lVal,
              (vaQuanity.lVal == 1) ? _T("") : _T("s"), vaPrice.fltVal);

    // construct an in memory image with the given text on it
    CComPtr<IPicture> sppict = makeReceiptBitmap(sz);

    // write the bitmap to the given stream; record the number of bytes written in m_nDocSize
    if (!sppict || FAILED(sppict->SaveAsFile(m_spstmOutput, TRUE, &m_nDocSize))) 
    {
        ATLTRACE("\nExit IBizTalkSerializer::AddDocument(), return E_FAIL\n");
        ATLASSERT(0);
        return E_FAIL;
    }
        
    // store the document handle for use by GetDocInfo()
    m_nDocHandle = nDocHandle;

    ATLTRACE(_T("\nExit IBizTalkSerializer::AddDocument([out] pdictTransport=%ls), return S_OK\n"), (BSTR)_TRACER(pdictTransport));
    return S_OK;
}

// IBizTalkSerializer::GetInterchangeInfo()
// ----------------------------------------
//  get information about the interchange.  This function is called after all documents
//  have been serialized.
//
//  on output,
//      pbstrInterchangeID = an identifer for the exchange to appear in the tracking database
//      pnNumGroups = number of groups in the interchange.  Must be 0 or 1 for BizTalk 2002.
//
//  return value
//      is S_OK on success and a negative HRESULT otherwise
//
STDMETHODIMP CFormPostSerializer::GetInterchangeInfo(BSTR * pbstrInterchangeID, long * pnNumGroups)
{
    ATLTRACE(_T("\nEnter IBizTalkSerializer::GetInterchangeInfo()\n"));
    if (!pbstrInterchangeID || !pnNumGroups) 
    {
        ATLTRACE("\nExit IBizTalkSerializer::GetInterchangeInfo(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }

    // allocate an id to go in the tracking database
    GUID guid;
    wchar_t wszGuid[128];
    if (FAILED(CoCreateGuid(&guid)))
    {
        ATLTRACE("\nExit IBizTalkSerializer::GetInterchangeInfo(), return E_FAIL\n");
        ATLASSERT(0);
        return E_FAIL;
    }
    StringFromGUID2(guid, wszGuid, sizeof wszGuid);
    *pbstrInterchangeID = CComBSTR(wszGuid).Detach();

    // set number of groups in the interchange (must be 0 or 1 for BizTalk 2002)
    *pnNumGroups = 0;

    ATLASSERT((*pnNumGroups == 0) || (*pnNumGroups == 1)); // required for BizTalk 2002
    ATLTRACE(_T("\nExit IBizTalkSerializer::GetInterchangeInfo([out] pbstrInterchangeID=\"%ls\", [out] pnNumGroups=%ld), return S_OK\n"), 
                    (BSTR)_TRACER(*pbstrInterchangeID), *pnNumGroups);
    return S_OK;
}

// IBizTalkSerializer::GetGroupInfo()
// ----------------------------------
//  get information about next group for the tracking database.
//
//  on output,
//      pnNumDocs = number of documents in the next group
//      pnGroupStartOffset = offset into the output stream for the start of the group
//                         = ofset to the first document in the group
//      pnGroupLen = total number of bytes in the group
//                 = total number of bytes in all documents in the group
//
//  return value
//      is S_OK on success and a negative HRESULT otherwise.
//      E_UNEXPECTED is returned if GetInterchangeInfo() specifies zero groups.
// 
STDMETHODIMP CFormPostSerializer::GetGroupInfo(long * pnNumDocs, LARGE_INTEGER * pnGroupStartOffset, long * pnGroupLen)
{
    ATLTRACE(_T("\nEnter IBizTalkSerializer::GetGroupInfo()\n"));
    if (!pnNumDocs || !pnGroupStartOffset || !pnGroupLen) 
    {
        ATLTRACE("\nExit IBizTalkSerializer::GetGroupInfo(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }
    ATLTRACE(_T("\nExit IBizTalkSerializer::GetGroupInfo([out] pnNumDocs=%ld, [out] pnGroupStartOffset=%I64d, [out] pnGroupLen), return E_NOTIMPL\n"), 
                    *pnNumDocs, pnGroupStartOffset->QuadPart, *pnGroupLen);
    return E_NOTIMPL;
}

// IBizTalkSerializer::GetDocInfo()
// --------------------------------
//  get information about next document for the tracking database.
//
//  on output,
//      pnDocHandle = handle for the next document in the group; this handle value
//                    was match one of the values passed into AddDocument()
//      pbSizeFromXMLDoc = offset into the output stream for the start of the group
//                         = ofset to the first document in the group
//      pbSizeFromXMLDoc = TRUE,  if BizTalk should infer the size and position.  In this case,
//                                the last 2 output parameters are not used.
//                         FALSE, if the size and position are specifed in the last 2 output parameters.
//
//                         Inference is done based on the XML returned from GetNextDocument().
//      pnDocStartOffset = the offset within the stream for the start of the document
//                         set only if pbSizeFromXMLDoc is set to FALSE
//      pnDocLen = the size within the stream for the start of the document's data in the unparsed form
//                    set only if pbSizeFromXMLDoc is set to FALSE
//
//  return value
//      is S_OK on success and a negative HRESULT otherwise
//
STDMETHODIMP CFormPostSerializer::GetDocInfo(long * pnDocHandle, BOOL * pbSizeFromXMLDoc, LARGE_INTEGER * pnDocStartOffset, long * pnDocLen)
{
    ATLTRACE(_T("\nEnter IBizTalkSerializer::GetDocInfo()\n"));
    if (!pnDocHandle || !pbSizeFromXMLDoc || !pnDocStartOffset || !pnDocLen) 
    {
        ATLTRACE("\nExit IBizTalkSerializer::GetDocInfo(), return E_POINTER\n");
        ATLASSERT(0);
        return E_POINTER;
    }

    // set the document handle and size to the value stored in AddDocument()
    *pnDocHandle = m_nDocHandle;
    *pnDocLen = m_nDocSize;

    // set the document offset which is 0 because there is only 1 document in the interchange    
    pnDocStartOffset->QuadPart = 0;

    // flag that we sent the size and offset here
    *pbSizeFromXMLDoc = FALSE;

    ATLTRACE(_T("\nExit IBizTalkSerializer::GetDocInfo([out] pnDocHandle=%ld, [out] pbSizeFromXMLDoc=%ld, [out] pnDocStartOffset=%I64u, [out] pnDocLen=%ld), return E_NOTIMPL\n"), *pnDocHandle, *pbSizeFromXMLDoc, pnDocStartOffset->QuadPart, *pnDocLen);
    return S_OK;;
}
